from torch_topological.datasets import SphereVsTorus

from torch_topological.nn.data import make_tensor

from torch_topological.nn import AlphaComplex

from torch.utils.data import DataLoader

batch_size = 64


class TestAlphaComplexBatchHandling:
    data_set = SphereVsTorus(n_point_clouds=3 * batch_size)
    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    ac = AlphaComplex()

    def test_processing(self):
        for (x, y) in self.loader:
            pers_info = self.ac(x)

            assert pers_info is not None
            assert len(pers_info) == batch_size

            pers_info_dense = make_tensor(pers_info)

            assert pers_info_dense is not None



from torch_topological.datasets import SphereVsTorus

from torch_topological.nn.data import batch_iter
from torch_topological.nn.data import make_tensor
from torch_topological.nn.data import PersistenceInformation

from torch_topological.nn import VietorisRipsComplex
from torch_topological.nn import WassersteinDistance

from torch import cdist
from torch import isinf
from torch.utils.data import DataLoader

import torch
import pytest

import numpy as np

batch_size = 64


class TestVietorisRipsComplex:
    data_set = SphereVsTorus(n_point_clouds=3 * batch_size)
    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    vr = VietorisRipsComplex(dim=1, p=1)

    def test_simple(self):
        for (x, y) in self.loader:
            pers_info = self.vr(x)

            assert pers_info is not None
            assert len(pers_info) == batch_size

    def test_predefined_distances(self):
        for (x, y) in self.loader:

            distances = cdist(x, x, p=1)
            pers_info1 = self.vr(distances, treat_as_distances=True)
            pers_info2 = self.vr(x, treat_as_distances=False)

            assert pers_info1 is not None
            assert pers_info2 is not None
            assert len(pers_info1) == batch_size
            assert len(pers_info1) == len(pers_info2)

            # Check that we are getting the same persistence diagrams.
            for pi1, pi2 in zip(pers_info1, pers_info2):
                dist = WassersteinDistance()(pi1, pi2)
                assert dist == pytest.approx(0.0)


class TestVietorisRipsComplexThreshold:
    data_set = SphereVsTorus(n_point_clouds=3 * batch_size)
    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    vr = VietorisRipsComplex(
        dim=1,
        p=1,
        threshold=0.1,
        keep_infinite_features=True
    )

    def test_threshold(self):
        for (x, y) in self.loader:
            pers_info = self.vr(x)

            assert pers_info is not None
            assert len(pers_info) == batch_size

            assert(torch.any(isinf(pers_info[0][0].diagram.flatten().sum())))


class TestVietorisRipsComplexBatchHandling:
    data_set = SphereVsTorus(n_point_clouds=3 * batch_size)
    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    vr = VietorisRipsComplex(dim=1)

    def test_processing(self):
        for (x, y) in self.loader:
            pers_info = self.vr(x)

            assert pers_info is not None
            assert len(pers_info) == batch_size

            pers_info_dense = make_tensor(pers_info)

            assert pers_info_dense is not None

    def test_ragged_processing(self):
        rng = np.random.default_rng()

        data = [
            np.random.default_rng().uniform(size=(rng.integers(32, 64), 8))
            for _ in range(batch_size)
        ]

        pers_info = self.vr(data)

        assert pers_info is not None
        assert len(pers_info) == batch_size

    def test_batch_iter(self):
        for (x, y) in self.loader:
            pers_info = self.vr(x)

            assert pers_info is not None
            assert len(pers_info) == batch_size

            # This is just to confirm that we can properly iterate over
            # this batch. Here, `batch_iter` is a little bit like `NoP`,
            # but in general, more complicated nested structures may be
            # present.
            assert sum(1 for x in batch_iter(pers_info)) == batch_size

            for x in batch_iter(pers_info):
                for y in x:
                    assert isinstance(y, PersistenceInformation)

            for x in batch_iter(pers_info, dim=0):

                # Make sure that we have something to iterate over.
                assert sum(1 for y in x) != 0

                for y in x:
                    assert isinstance(y, PersistenceInformation)
                    assert y.dimension == 0



import numpy as np

import torch

from torch_topological.nn import PersistenceInformation
from torch_topological.nn import WassersteinDistance


def wrap(diagram):
    diagram = torch.as_tensor(diagram, dtype=torch.float)
    return [
        PersistenceInformation([], diagram)
    ]


class TestWassersteinDistance:

    def test_simple(self):
        X = [
            (3, 5),
            (1, 2),
            (3, 4)
        ]

        Y = [
            (0, 0),
            (3, 4),
            (5, 3),
            (7, 4)
        ]

        X = wrap(X)
        Y = wrap(Y)

        dist = WassersteinDistance()(X, Y)
        assert dist > 0.0

    def test_random(self):
        n_points = 10
        n_instances = 10

        for i in range(n_instances):
            X = np.random.default_rng().normal(size=(n_points, 2))
            Y = np.random.default_rng().normal(size=(n_points, 2))

            X = wrap(X)
            Y = wrap(Y)

            dist = WassersteinDistance()(X, Y)
            assert dist > 0.0

    def test_almost_zero(self):
        n_points = 100

        X = np.random.default_rng().uniform(-1e-16, 1e-11, size=(n_points, 2))
        Y = np.random.default_rng().uniform(-1e-16, 1e-11, size=(n_points, 2))

        X = wrap(X)
        Y = wrap(Y)

        dist = WassersteinDistance()(X, Y)
        assert dist > 0.0



from torch_topological.datasets import SphereVsTorus

from torch_topological.nn.data import make_tensor

from torch_topological.nn import VietorisRipsComplex
from torch_topological.nn.layers import StructureElementLayer

from torch.utils.data import DataLoader

batch_size = 32


class TestStructureElementLayer:
    data_set = SphereVsTorus(n_point_clouds=2 * batch_size)
    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    vr = VietorisRipsComplex(dim=1)

    layer = StructureElementLayer(10)

    def test_processing(self):
        for (x, y) in self.loader:
            pers_info = make_tensor(self.vr(x))
            output = self.layer(pers_info)

            assert pers_info is not None
            assert output is not None



from torch_topological.data import sample_from_unit_cube

from torch_topological.nn import MultiScaleKernel
from torch_topological.nn import VietorisRipsComplex


class TestMultiScaleKernel:
    vr = VietorisRipsComplex(dim=1)
    kernel = MultiScaleKernel(1.0)
    X = sample_from_unit_cube(100)
    Y = sample_from_unit_cube(100)

    def test_pseudo_metric(self):
        pers_info_X, pers_info_Y = self.vr([self.X, self.Y])
        k_XX = self.kernel(pers_info_X, pers_info_X)
        k_YY = self.kernel(pers_info_Y, pers_info_Y)
        k_XY = self.kernel(pers_info_X, pers_info_Y)

        assert k_XY > 0
        assert k_XX > 0
        assert k_YY > 0

        assert k_XX + k_YY - 2 * k_XY >= 0



from torchvision.datasets import MNIST

from torchvision.transforms import Compose
from torchvision.transforms import Normalize
from torchvision.transforms import ToTensor

from torch_topological.nn import CubicalComplex

from torch.utils.data import Dataset
from torch.utils.data import DataLoader
from torch.utils.data import RandomSampler

from torch_topological.nn.data import batch_iter
from torch_topological.nn.data import make_tensor
from torch_topological.nn.data import PersistenceInformation

import numpy as np

import torch


class RandomDataSet(Dataset):
    def __init__(self, n_samples, dim, side_length, n_channels):
        self.dim = dim
        self.side_length = side_length
        self.n_samples = n_samples
        self.n_channels = n_channels

        self.data = np.random.default_rng().normal(
            size=(n_samples, n_channels, *([side_length] * dim))
        )

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)


class TestCubicalComplex:
    batch_size = 32

    def test_single_image(self):
        x = np.random.default_rng().normal(size=(32, 32))
        x = torch.as_tensor(x)
        cc = CubicalComplex()

        pers_info = cc(x)

        assert pers_info is not None
        assert len(pers_info) == 2
        assert pers_info[0].dimension == 0
        assert pers_info[1].dimension == 1

    def test_image_with_channels(self):
        x = np.random.default_rng().normal(size=(3, 32, 32))
        x = torch.as_tensor(x)
        cc = CubicalComplex()

        pers_info = cc(x)

        assert pers_info is not None
        assert len(pers_info) == 3
        assert len(pers_info[0]) == 2
        assert pers_info[0][0].dimension == 0
        assert pers_info[1][1].dimension == 1

    def test_image_with_channels_and_batch(self):
        x = np.random.default_rng().normal(size=(self.batch_size, 3, 32, 32))
        x = torch.as_tensor(x)
        cc = CubicalComplex()

        pers_info = cc(x)

        assert pers_info is not None
        assert len(pers_info) == self.batch_size
        assert len(pers_info[0][0]) == 2
        assert pers_info[0][0][0].dimension == 0
        assert pers_info[1][1][1].dimension == 1

    def test_2d(self):
        for n_channels in [1, 3]:
            for squeeze in [False, True]:
                data_set = RandomDataSet(128, 2, 8, n_channels)
                loader = DataLoader(
                    data_set,
                    self.batch_size,
                    shuffle=True,
                    drop_last=False
                )

                if squeeze:
                    data_set.data = data_set.data.squeeze()

                cc = CubicalComplex(dim=2)

                for batch in loader:
                    pers_info = cc(batch)

                    assert pers_info is not None

    def test_3d(self):
        data_set = RandomDataSet(128, 3, 8, 1)
        loader = DataLoader(
            data_set,
            self.batch_size,
            shuffle=True,
            drop_last=False
        )

        cc = CubicalComplex(dim=3)

        for batch in loader:
            pers_info = cc(batch)

            assert pers_info is not None
            assert len(pers_info) == self.batch_size


class TestCubicalComplexBatchHandling:
    batch_size = 64

    data_set = MNIST(
        './data/MNIST',
        download=True,
        train=False,
        transform=Compose(
            [
                ToTensor(),
                Normalize([0.5], [0.5])
            ]
        ),
    )

    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=False,
        drop_last=True,
        sampler=RandomSampler(
            data_set,
            replacement=True,
            num_samples=100
        )
    )

    cc = CubicalComplex()

    def test_processing(self):
        for (x, y) in self.loader:
            pers_info = self.cc(x)

            assert pers_info is not None
            assert len(pers_info) == self.batch_size

            pers_info_dense = make_tensor(pers_info)

            assert pers_info_dense is not None

    def test_batch_iter(self):
        for (x, y) in self.loader:
            pers_info = self.cc(x)

            assert pers_info is not None
            assert len(pers_info) == self.batch_size

            assert sum(1 for x in batch_iter(pers_info)) == self.batch_size

            for x in batch_iter(pers_info):
                for y in x:
                    assert isinstance(y, PersistenceInformation)

            for x in batch_iter(pers_info, dim=0):

                # Make sure that we have something to iterate over.
                assert sum(1 for y in x) != 0

                for y in x:
                    assert isinstance(y, PersistenceInformation)
                    assert y.dimension == 0
